/* buffer.c
 *   I/O buffering system
 *   author: Frank Barrus
 */


#include "types.h"
#include "error.h"


#define BNODE_SIZE	64		/* size of a buffer node - 
					   should be a power of two <= 256 */
#define BNODE_CHARS	(BNODE_SIZE-2)	/* number of characters in bnode */


typedef struct bnode {
	uint16		next;		/* index of next buffer */
	char		str[BNODE_CHARS]; 
					/* actual buffered characters */
} BNode;



typedef struct io_buf {
	uint16		r_buf;		/* index of first buf for reading */
	uint16		w_buf;		/* index of last bug for writing */
	uint8		r_ofs;		/* offset into first buf for reading */
	uint8		w_ofs;		/* offset into last buf for writing */
	uint32		bufsize;	/* current size of buffer */
	uint32		owner;		/* owner id or 0 for none */
} IO_Buf;



/* global variables to keep track of buffer system:   
 */

BNode	*bnode;			/* pointer to array of bnodes -
				   shifted so indexing starts at 1 */
uint16	num_bnode;		/* number of bnodes */
uint16 	free_bnode;		/* index of first free buffer */
IO_Buf	*io_buf;		/* pointer to array of I/O buffers */
uint16	num_io_buf;		/* number of I/O buffers */



/* initialize the buffering system using the supplied block of memory
 * to store buffers:
 */
void buf_init(ptr,size,numbuf)
	char *ptr;
	uint32 size;	
	int numbuf;
{
int i;
	/* set up arrays of buffers */
	num_io_buf = numbuf;
	io_buf = (IO_Buf*)ptr;
	ptr += (sizeof(IO_Buf) * numbuf);
	bnode = (BNode*)((uint32)ptr - BNODE_SIZE);
	num_bnode = (size - sizeof(IO_Buf)*numbuf) / BNODE_SIZE;

	/* initialize free list */
	for(i=1; i<num_bnode; i++)
		bnode[i].next = i+1;
	for(i=0; i<num_io_buf; i++)
		io_buf[i].owner = 0;
	bnode[num_bnode].next = 0;
	free_bnode = 1;
}



/* open an available buffer, assign it the given I/O control
 * functions, and return its id: 
 */
int buf_open()
{
int i;
IO_Buf *io;
	/* find a free I/O buffer: */
	i = 0;
	while(i < num_io_buf && io_buf[i].owner)
		i++;
	if (i == num_io_buf)
		return(ERR_NOBUF);

	/* initialize it */
	io = &io_buf[i];
	io->owner = getpid();
	io->r_buf = 0;
	io->w_buf = 0;
	io->r_ofs = 0;
	io->w_ofs = BNODE_CHARS;
	io->bufsize = 0;
	return(i);
}



/* release a specified buffer: 
 */
status buf_close(b)
	int b;
{
IO_Buf *io;
	io = &io_buf[b];
	if (!io->owner)
		return(ERR_INVBUF);
	buf_purge(b);
	io->owner = 0;
	return(OK);
}



/* block a process until all data is flushed - not implemented yet
 */
status buf_flush(b)
	int b;
{
	return(OK);
}


/* purge all data from a specified buffer: 
 */
status buf_purge(b)
	int b;
{
int i, n;	/* bnode indices */
IO_Buf *io;
	io = &io_buf[b];
	if (!io->owner)
		return(ERR_INVBUF);
	i = io->r_buf;
	while(i) {
		n = bnode[i].next;
		bnode[i].next = free_bnode;
		free_bnode = i;
		i = n;
	}
	io->bufsize = 0;
	io->r_buf = 0;
	io->w_buf = 0;
	io->r_ofs = 0;
	io->w_ofs = BNODE_CHARS;
	return(OK);
}



int buf_read(b, ptr, n)
	int b;
	char *ptr;
	uint32 n;
{
IO_Buf *io;
uint16 bnext;	/* index of next bnode */
int i;		/* number of characters transferred */
	io = &io_buf[b];
	if (!io->owner || (io->bufsize && !io->r_buf))
		return(ERR_INVBUF);
	i = 0;
	while(io->bufsize && n--) {
		i++;
		if(io->r_ofs >= BNODE_CHARS) {
			bnext = bnode[io->r_buf].next;
			bnode[io->r_buf].next = free_bnode;
			free_bnode = io->r_buf;
			io->r_buf = bnext;
			io->r_ofs = 0;
		}
		*ptr++ = bnode[io->r_buf].str[io->r_ofs++];
		io->bufsize--;
	}
	return(i);
}



int buf_write(b, ptr, n)
	int b;
	char *ptr;
	uint32 n;
{
int i;		/* number of characters actually written */
uint16 bnext;	/* index of next bnode */
IO_Buf *io;
	io = &io_buf[b];
/*	cprintf("b=%d w_ofs=%d r_ofs=%d w_buf=%d r_buf=%d\r\n",
		b, io->w_ofs, io->r_ofs, io->w_buf, io->r_buf);  
*/
	if (!io->owner || (io->bufsize && !io->w_buf))
		return(ERR_INVBUF);
	i = 0;
	while(n--) {
		if (io->w_ofs >= BNODE_CHARS) {
			if (!(bnext = free_bnode))
				return(i);
			free_bnode = bnode[free_bnode].next;
			if (io->w_buf)
				bnode[io->w_buf].next = bnext;
			else
				io->r_buf = bnext;
			bnode[bnext].next = 0;
			io->w_buf = bnext;
			io->w_ofs = 0;
		}
		bnode[io->w_buf].str[io->w_ofs++] = *ptr++;
		io->bufsize++;
		i++;
	}
	return(i);
}



int buf_getsize(b)
	int b;
{
IO_Buf *io;
	io = &io_buf[b];
	if (io->owner)
		return(io->bufsize);
	else
		return(ERR_INVBUF);
}


buf_debug()
{
int i;
int x;
	printf("freebuf: %d (%d buffers)\n", free_bnode, num_bnode);
	for(i=0; i<10; i++) {
		printf("%d -> %d: ", i, bnode[i].next);
		for(x=0; x<62; x++)
			printf("%2x ", bnode[i].str[x]);
		printf("\n");
	}
}
